@@ -36,7 +36,9 @@ typedef signed char         int8_t;   // unqualified 'char' is unsigned on ARM
  * This prevents GCC from emitting GOT based relocations, and use R_ARM_REL32
  * relative relocations instead, which are more suitable for static binaries.
  */
+#ifdef __GNUC__
 #pragma GCC visibility push (hidden)
+#endif
 
 //
 // Basic EFI types of various widths
@@ -121,9 +123,13 @@ typedef uint32_t   UINTN;
 
 //
 // When build similiar to FW, then link everything together as
-// one big module.
+// one big module. For the MSVC toolchain, we simply tell the
+// linker what our driver init function is using /ENTRY.
 //
-
+#if defined(_MSC_EXTENSIONS)
+#define EFI_DRIVER_ENTRY_POINT(InitFunction) \
+    __pragma(comment(linker, "/ENTRY:" # InitFunction))
+#else
 #define EFI_DRIVER_ENTRY_POINT(InitFunction)    \
     UINTN                                       \
     InitializeDriver (                          \
@@ -140,6 +146,7 @@ typedef uint32_t   UINTN;
         EFI_SYSTEM_TABLE *systab                \
         ) __attribute__((weak,                  \
                 alias ("InitializeDriver")));
+#endif
 
 #define LOAD_INTERNAL_DRIVER(_if, type, name, entry)    \
         (_if)->LoadInternal(type, name, entry)
